# Fixtures for doctests
from pathlib import Path
import pytest
from time import perf_counter
from datetime import timedelta
import os
import sys
from pytest_pretty import CustomTerminalReporter

from _pytest.terminal import TerminalReporter
from _pytest.pathlib import bestrelpath


class SKTerminalReporter(CustomTerminalReporter):
    """Custom terminal reporter to display test runtimes.

    It appends the cumulative runtime each time a new file is tested.
    """

    currentfspath: Path | None
    _start_time: float | None

    def write_fspath_result(self, nodeid: str, res, **markup: bool) -> None:
        if getattr(self, '_start_time', None) is None:
            self._start_time = perf_counter()

        fspath = self.config.rootpath / nodeid.split('::')[0]
        if fspath != self.currentfspath:
            if self.currentfspath is not None and self._show_progress_info:
                # call method to write information about progress
                # padding spaces and percentage information
                self._write_progress_information_filling_space()
                if os.environ.get('CI', False):
                    # write time elapsed since the beginning of the test suite
                    elapsed = timedelta(seconds=int(perf_counter() - self._start_time))
                    self.write(f' [{elapsed}]')

            self.currentfspath = fspath
            relfspath = bestrelpath(self.startpath, fspath)
            self._tw.line()
            self.write(relfspath + ' ')

        self.write(res, flush=True, **markup)

    def short_test_summary(self):
        # Don't use table-based summary from pytest_pretty,
        # use pytest's original one that won't truncate long file and test names
        TerminalReporter.short_test_summary(self)


@pytest.hookimpl(trylast=True)
def pytest_configure(config: pytest.Config) -> None:
    # Get the standard terminal reporter plugin and replace it with ours
    standard_reporter = config.pluginmanager.getplugin('terminalreporter')
    custom_reporter = SKTerminalReporter(config, sys.stdout)
    if standard_reporter._session is not None:
        # if session is already set we need to copy them from
        # the previous reporter
        custom_reporter._session = standard_reporter._session
    config.pluginmanager.unregister(standard_reporter)
    config.pluginmanager.register(custom_reporter, 'terminalreporter')



"""
=====================
Image Deconvolution
=====================
In this example, we deconvolve an image using the
Richardson–Lucy algorithm ([1]_, [2]_, [3]_).

The algorithm is based on a point spread function (PSF),
described as the impulse response of the
optical system. The blurred image is sharpened through a number of
iterations, which needs to be hand-tuned.

.. [1] William Hadley Richardson, "Bayesian-Based Iterative
       Method of Image Restoration",
       J. Opt. Soc. Am. A 27, 1593-1607 (1972), :DOI:`10.1364/JOSA.62.000055`

.. [2] https://en.wikipedia.org/wiki/Richardson%E2%80%93Lucy_deconvolution

.. [3] https://www.strollswithmydog.com/richardson-lucy-algorithm/
"""

import numpy as np
import matplotlib.pyplot as plt
import skimage as ski

from scipy.signal import convolve2d as conv2


rng = np.random.default_rng()

# Convert astronaut image to grayscale
astro = ski.color.rgb2gray(ski.data.astronaut())

# Define PSF
psf = np.ones((5, 5)) / 25

# Convolve image with the PSF to simulate a blurred image
astro_blurred = conv2(astro, psf, 'same')

# Add Poisson noise to the blurred image (https://en.wikipedia.org/wiki/Shot_noise)
max_photon_count = 1000
astro_noisy = rng.poisson(astro_blurred * max_photon_count) / max_photon_count

# Normalize noisy image
astro_noisy /= np.max(astro_noisy)

# Restore image by means of deconvolution
deconvolved_RL = ski.restoration.richardson_lucy(astro_noisy, psf, num_iter=30)

fig, ax = plt.subplots(ncols=3, figsize=(8, 5))
plt.gray()

for a in (ax[0], ax[1], ax[2]):
    a.axis('off')

ax[0].imshow(astro)
ax[0].set_title('Original Data')

ax[1].imshow(astro_noisy)
ax[1].set_title('Noisy data')

ax[2].imshow(deconvolved_RL)
ax[2].set_title('Restoration using\nRichardson-Lucy')


fig.subplots_adjust(wspace=0.02, hspace=0.2, top=0.9, bottom=0.05, left=0, right=1)
plt.show()



from itertools import product
import re
import textwrap

import numpy as np
import pytest
from numpy.testing import assert_almost_equal, assert_array_almost_equal, assert_equal

from skimage.transform import (
    AffineTransform,
    EssentialMatrixTransform,
    EuclideanTransform,
    FundamentalMatrixTransform,
    PiecewiseAffineTransform,
    PolynomialTransform,
    ProjectiveTransform,
    SimilarityTransform,
    estimate_transform,
    matrix_transform,
)
from skimage.transform._geometric import (
    _GeometricTransform,
    _affine_matrix_from_vector,
    _calc_center_normalize,
    _center_and_normalize_points,
    _apply_homogeneous,
    _euler_rotation_matrix,
    _append_homogeneous_dim,
    TRANSFORMS,
)
from skimage import data

from skimage._shared.utils import FailedEstimationAccessError

SRC = np.array(
    [
        [-12.3705, -10.5075],
        [-10.7865, 15.4305],
        [8.6985, 10.8675],
        [11.4975, -9.5715],
        [7.8435, 7.4835],
        [-5.3325, 6.5025],
        [6.7905, -6.3765],
        [-6.1695, -0.8235],
    ]
)
DST = np.array(
    [
        [0, 0],
        [0, 5800],
        [4900, 5800],
        [4900, 0],
        [4479, 4580],
        [1176, 3660],
        [3754, 790],
        [1024, 1931],
    ]
)

# Transforms accepting homogeneous matrix as input.
HMAT_TFORMS = (
    FundamentalMatrixTransform,
    ProjectiveTransform,
    AffineTransform,
    EuclideanTransform,
    SimilarityTransform,
)

# Transforms allowing ND matrix.
HMAT_TFORMS_ND = (
    ProjectiveTransform,
    AffineTransform,
    EuclideanTransform,
    SimilarityTransform,
)


def test_estimate_transform():
    for tform in ('euclidean', 'similarity', 'affine', 'projective', 'polynomial'):
        estimate_transform(tform, SRC[:2, :], DST[:2, :])
    with pytest.raises(ValueError):
        estimate_transform('foobar', SRC[:2, :], DST[:2, :])


def test_matrix_transform():
    tform = AffineTransform(scale=(0.1, 0.5), rotation=2)
    assert_equal(tform(SRC), matrix_transform(SRC, tform.params))


def test_euclidean_estimation():
    # exact solution
    tform = estimate_transform('euclidean', SRC[:2, :], SRC[:2, :] + 10)
    assert_almost_equal(tform(SRC[:2, :]), SRC[:2, :] + 10)
    assert_almost_equal(tform.params[0, 0], tform.params[1, 1])
    assert_almost_equal(tform.params[0, 1], -tform.params[1, 0])

    # over-determined
    tform2 = estimate_transform('euclidean', SRC, DST)
    assert_almost_equal(tform2.inverse(tform2(SRC)), SRC)
    assert_almost_equal(tform2.params[0, 0], tform2.params[1, 1])
    assert_almost_equal(tform2.params[0, 1], -tform2.params[1, 0])

    # via from_estimate classmethod
    tform3 = EuclideanTransform.from_estimate(SRC, DST)
    assert_almost_equal(tform3.params, tform2.params)
    # via estimate method
    tform4 = EuclideanTransform.identity()
    with pytest.warns(FutureWarning, match='`estimate` is deprecated'):
        assert tform4.estimate(SRC, DST)
    assert_almost_equal(tform4.params, tform2.params)


@pytest.mark.parametrize(
    'tform_class, has_scale', ((EuclideanTransform, False), (SimilarityTransform, True))
)
def test_3d_euclidean_similarity_estimation(tform_class, has_scale):
    src_points = np.random.rand(1000, 3)

    # Random transformation for testing
    angles = np.random.random((3,)) * 2 * np.pi - np.pi
    rotation_matrix = _euler_rotation_matrix(angles)
    if has_scale:
        scale = np.random.randint(0, 20)
        rotation_matrix *= scale
    translation_vector = np.random.random((3,))
    dst_points = []
    for pt in src_points:
        pt_r = pt.reshape(3, 1)
        dst = np.matmul(rotation_matrix, pt_r) + translation_vector.reshape(3, 1)
        dst = dst.reshape(3)
        dst_points.append(dst)

    dst_points = np.array(dst_points)
    # estimating the transformation
    tform = tform_class.from_estimate(src_points, dst_points)
    assert tform
    assert_almost_equal(tform.rotation, rotation_matrix)
    assert_almost_equal(tform.translation, translation_vector)
    if has_scale:
        assert_almost_equal(tform.scale, scale)
    # estimate method
    tform2 = tform_class.identity(3)
    with pytest.warns(FutureWarning, match='`estimate` is deprecated'):
        assert tform2.estimate(src_points, dst_points)
    assert_equal(tform.params, tform2.params)


def test_euclidean_init():
    # init with implicit parameters
    rotation = 1
    translation = (1, 1)
    tform = EuclideanTransform(rotation=rotation, translation=translation)
    assert_almost_equal(tform.rotation, rotation)
    assert_almost_equal(tform.translation, translation)

    # init with transformation matrix
    tform2 = EuclideanTransform(tform.params)
    assert_almost_equal(tform2.rotation, rotation)
    assert_almost_equal(tform2.translation, translation)

    # test special case for scale if rotation=0
    rotation = 0
    translation = (1, 1)
    tform = EuclideanTransform(rotation=rotation, translation=translation)
    assert_almost_equal(tform.rotation, rotation)
    assert_almost_equal(tform.translation, translation)

    # test special case for scale if rotation=90deg
    rotation = np.pi / 2
    translation = (1, 1)
    tform = EuclideanTransform(rotation=rotation, translation=translation)
    assert_almost_equal(tform.rotation, rotation)
    assert_almost_equal(tform.translation, translation)


def test_similarity_estimation():
    # exact solution
    tform = estimate_transform('similarity', SRC[:2, :], DST[:2, :])
    assert_almost_equal(tform(SRC[:2, :]), DST[:2, :])
    assert_almost_equal(tform.params[0, 0], tform.params[1, 1])
    assert_almost_equal(tform.params[0, 1], -tform.params[1, 0])

    # over-determined
    tform2 = estimate_transform('similarity', SRC, DST)
    assert_almost_equal(tform2.inverse(tform2(SRC)), SRC)
    assert_almost_equal(tform2.params[0, 0], tform2.params[1, 1])
    assert_almost_equal(tform2.params[0, 1], -tform2.params[1, 0])

    # via from_estimate classmethod
    tform3 = SimilarityTransform.from_estimate(SRC, DST)
    assert_almost_equal(tform3.params, tform2.params)
    # via estimate method
    tform4 = SimilarityTransform()
    with pytest.warns(FutureWarning, match='`estimate` is deprecated'):
        assert tform4.estimate(SRC, DST)
    assert_almost_equal(tform4.params, tform2.params)


def test_similarity_init():
    # init with implicit parameters
    scale = 0.1
    rotation = 1
    translation = (1, 1)
    tform = SimilarityTransform(scale=scale, rotation=rotation, translation=translation)
    assert_almost_equal(tform.scale, scale)
    assert_almost_equal(tform.rotation, rotation)
    assert_almost_equal(tform.translation, translation)

    # init with transformation matrix
    tform2 = SimilarityTransform(tform.params)
    assert_almost_equal(tform2.scale, scale)
    assert_almost_equal(tform2.rotation, rotation)
    assert_almost_equal(tform2.translation, translation)

    # test special case for scale if rotation=0
    scale = 0.1
    rotation = 0
    translation = (1, 1)
    tform = SimilarityTransform(scale=scale, rotation=rotation, translation=translation)
    assert_almost_equal(tform.scale, scale)
    assert_almost_equal(tform.rotation, rotation)
    assert_almost_equal(tform.translation, translation)

    # test special case for scale if rotation=90deg
    scale = 0.1
    rotation = np.pi / 2
    translation = (1, 1)
    tform = SimilarityTransform(scale=scale, rotation=rotation, translation=translation)
    assert_almost_equal(tform.scale, scale)
    assert_almost_equal(tform.rotation, rotation)
    assert_almost_equal(tform.translation, translation)

    # test special case for scale where the rotation isn't exactly 90deg,
    # but very close
    scale = 1.0
    rotation = np.pi / 2
    translation = (0, 0)
    params = np.array(
        [[0, -1, 1.33226763e-15], [1, 2.22044605e-16, -1.33226763e-15], [0, 0, 1]]
    )
    tform = SimilarityTransform(params)
    assert_almost_equal(tform.scale, scale)
    assert_almost_equal(tform.rotation, rotation)
    assert_almost_equal(tform.translation, translation)

    # With scalar scale and 3D, we get a FutureWarning.  This is to
    # generalize the rule that dimensionality should be implied by
    # input parameters, when given.
    with pytest.warns(FutureWarning):
        tf = SimilarityTransform(scale=4, dimensionality=3)
    assert_equal(tf([[1, 1, 1]]), [[4, 4, 4]])
    # Not so if we specify some other input giving dimensionality.
    tf = SimilarityTransform(scale=4, translation=(0, 0, 0))
    assert_equal(tf([[1, 1, 1]]), [[4, 4, 4]])
    # Or if we are in 2D, by analogy to shear etc - scalar implies 2D.
    tf = SimilarityTransform(scale=4)
    assert_equal(tf([[1, 1]]), [[4, 4]])


def test_affine_estimation():
    # exact solution
    tform = estimate_transform('affine', SRC[:3, :], DST[:3, :])
    assert_almost_equal(tform(SRC[:3, :]), DST[:3, :])

    # over-determined
    tform2 = estimate_transform('affine', SRC, DST)
    assert_almost_equal(tform2.inverse(tform2(SRC)), SRC)

    # via from_estimate classmethod
    tform3 = AffineTransform.from_estimate(SRC, DST)
    assert_almost_equal(tform3.params, tform2.params)
    # via estimate method
    tform4 = AffineTransform.identity()
    with pytest.warns(FutureWarning, match='`estimate` is deprecated'):
        assert tform4.estimate(SRC, DST)
    assert_almost_equal(tform4.params, tform2.params)


def test_affine_init():
    # init with implicit parameters
    scale = (0.1, 0.13)
    rotation = 1
    shear = 0.1
    translation = (1, 1)
    tform = AffineTransform(
        scale=scale, rotation=rotation, shear=shear, translation=translation
    )
    assert_almost_equal(tform.scale, scale)
    assert_almost_equal(tform.rotation, rotation)
    assert_almost_equal(tform.shear, shear)
    assert_almost_equal(tform.translation, translation)

    # init with transformation matrix
    tform2 = AffineTransform(tform.params)
    assert_almost_equal(tform2.scale, scale)
    assert_almost_equal(tform2.rotation, rotation)
    assert_almost_equal(tform2.shear, shear)
    assert_almost_equal(tform2.translation, translation)

    # scalar vs. tuple scale arguments
    assert_almost_equal(
        AffineTransform(scale=0.5).scale, AffineTransform(scale=(0.5, 0.5)).scale
    )


def test_affine_shear():
    shear = 0.1
    # expected horizontal shear transform
    cx = -np.tan(shear)
    expected = np.array([[1, cx, 0], [0, 1, 0], [0, 0, 1]])

    tform = AffineTransform(shear=shear)
    assert_almost_equal(tform.params, expected)

    shear = (1.2, 0.8)
    # expected x, y shear transform
    cx = -np.tan(shear[0])
    cy = -np.tan(shear[1])
    expected = np.array([[1, cx, 0], [cy, 1, 0], [0, 0, 1]])

    tform = AffineTransform(shear=shear)
    assert_almost_equal(tform.params, expected)


@pytest.mark.parametrize(
    'pts, params',
    product(
        (SRC, DST),
        (
            dict(scale=(4, 5), shear=(1.4, 1.8), rotation=0.4, translation=(10, 12)),
            dict(
                scale=(-0.5, 3), shear=(-0.3, -0.1), rotation=1.4, translation=(-4, 3)
            ),
        ),
    ),
)
def test_affine_params(pts, params):
    # Test AffineTransform against docstring algorithm.
    out = AffineTransform(**params)(pts)
    docstr_out = _apply_aff_2d(pts, **params)
    assert np.allclose(out, docstr_out)


def _apply_aff_2d(pts, scale, rotation, shear, translation):
    # Algorithm from AffineTransform docstring.
    x, y = pts.T
    sx, sy = scale
    shear_x, shear_y = shear
    translation_x, translation_y = translation
    cos, tan, sin = np.cos, np.tan, np.sin
    X = (
        sx * x * (cos(rotation) + tan(shear_y) * sin(rotation))
        - sy * y * (tan(shear_x) * cos(rotation) + sin(rotation))
        + translation_x
    )
    Y = (
        sx * x * (sin(rotation) - tan(shear_y) * cos(rotation))
        - sy * y * (tan(shear_x) * sin(rotation) - cos(rotation))
        + translation_y
    )
    return np.stack((X, Y), axis=1)


def test_piecewise_affine():
    tform = PiecewiseAffineTransform.from_estimate(SRC, DST)
    assert tform
    # make sure each single affine transform is exactly estimated
    assert_almost_equal(tform(SRC), DST)
    assert_almost_equal(tform.inverse(DST), SRC)
    # via estimate method
    tform2 = PiecewiseAffineTransform()
    with pytest.warns(FutureWarning, match='`estimate` is deprecated'):
        assert tform2.estimate(SRC, DST)
    assert_almost_equal(tform2(SRC), DST)
    assert_almost_equal(tform2.inverse(DST), SRC)


def test_fundamental_matrix_estimation():
    src = np.array(
        [
            1.839035,
            1.924743,
            0.543582,
            0.375221,
            0.473240,
            0.142522,
            0.964910,
            0.598376,
            0.102388,
            0.140092,
            15.994343,
            9.622164,
            0.285901,
            0.430055,
            0.091150,
            0.254594,
        ]
    ).reshape(-1, 2)
    dst = np.array(
        [
            1.002114,
            1.129644,
            1.521742,
            1.846002,
            1.084332,
            0.275134,
            0.293328,
            0.588992,
            0.839509,
            0.087290,
            1.779735,
            1.116857,
            0.878616,
            0.602447,
            0.642616,
            1.028681,
        ]
    ).reshape(-1, 2)

    tform = estimate_transform('fundamental', src, dst)

    # Reference values obtained using COLMAP SfM library.
    tform_ref = np.array(
        [
            [-0.217859, 0.419282, -0.0343075],
            [-0.0717941, 0.0451643, 0.0216073],
            [0.248062, -0.429478, 0.0221019],
        ]
    )
    assert_almost_equal(tform.params, tform_ref, 6)


def test_fundamental_matrix_residuals():
    essential_matrix_tform = EssentialMatrixTransform(
        rotation=np.eye(3), translation=np.array([1, 0, 0])
    )
    tform = FundamentalMatrixTransform()
    tform.params = essential_matrix_tform.params
    src = np.array([[0, 0], [0, 0], [0, 0]])
    dst = np.array([[2, 0], [2, 1], [2, 2]])
    assert_almost_equal(tform.residuals(src, dst) ** 2, [0, 0.5, 2])


@pytest.mark.parametrize('array_like_input', [False, True])
def test_fundamental_matrix_forward(array_like_input):
    if array_like_input:
        rotation = [[1, 0, 0], [0, 1, 0], [0, 0, 1]]
        translation = (1, 0, 0)
    else:
        rotation = np.eye(3)
        translation = np.array([1, 0, 0])
    essential_matrix_tform = EssentialMatrixTransform(
        rotation=rotation, translation=translation
    )
    if array_like_input:
        params = [list(p) for p in essential_matrix_tform.params]
    else:
        params = essential_matrix_tform.params
    tform = FundamentalMatrixTransform(matrix=params)
    src = np.array([[0, 0], [0, 1], [1, 1]])
    assert_almost_equal(tform(src), [[0, -1, 0], [0, -1, 1], [0, -1, 1]])


def test_fundamental_matrix_inverse():
    essential_matrix_tform = EssentialMatrixTransform(
        rotation=np.eye(3), translation=np.array([1, 0, 0])
    )
    tform = FundamentalMatrixTransform()
    tform.params = essential_matrix_tform.params
    src = np.array([[0, 0], [0, 1], [1, 1]])
    assert_almost_equal(tform.inverse(src), [[0, 1, 0], [0, 1, -1], [0, 1, -1]])


def test_fundamental_matrix_inverse_estimation():
    src = np.array(
        [
            1.839035,
            1.924743,
            0.543582,
            0.375221,
            0.473240,
            0.142522,
            0.964910,
            0.598376,
            0.102388,
            0.140092,
            15.994343,
            9.622164,
            0.285901,
            0.430055,
            0.091150,
            0.254594,
        ]
    ).reshape(-1, 2)

    dst = np.array(
        [
            1.002114,
            1.129644,
            1.521742,
            1.846002,
            1.084332,
            0.275134,
            0.293328,
            0.588992,
            0.839509,
            0.087290,
            1.779735,
            1.116857,
            0.878616,
            0.602447,
            0.642616,
            1.028681,
        ]
    ).reshape(-1, 2)

    # Inverse of (src -> dst) transform should be equivalent to
    # (dst -> src) transformation
    tform = estimate_transform('fundamental', src, dst)
    tform_inv = estimate_transform('fundamental', dst, src)

    np.testing.assert_array_almost_equal(tform.inverse.params, tform_inv.params)


def _calc_distances(src, dst, F, metric='distance'):
    """Distances between calculated epipolar lines and points

    Parameters
    ----------
    src : (N, D) array
        Points in first image.
    dst : (N, D) array
        Matching points in second image.
    F : (3, 3) array
        Fundamental matrix mapping `src` to epipolar lines passing through `dst`.
    metric : {'distance', 'epip-distances'}, optional
        Matrix for distance between actual points `dst` and epipolar lines
        generated from `F`.  'distance' is signed distance from [1]_.
        'epip-distances' is the squared sum of distances of points from epipolar
        lines in both images.  See [2]_, section 7.1.4.

    Notes
    -----
    See `Wikipedia on point-line distance
    <https://en.wikipedia.org/wiki/Distance_from_a_point_to_a_line#A_vector_projection_proof>`__
    for standard distance formula, and various proofs.

    References
    ----------
    .. [1] Zhang, Zhengyou. "Determining the epipolar geometry and its
           uncertainty: A review." International journal of computer vision 27
           (1998): 161-195.
           :DOI:`10.1023/A:1007941100561`
           https://www.microsoft.com/en-us/research/wp-content/uploads/2016/11/RR-2927.pdf
    .. [2] Hartley, Richard I. "In defense of the eight-point algorithm."
           Pattern Analysis and Machine Intelligence, IEEE Transactions on 19.6
           (1997): 580-593.
           https://users.cecs.anu.edu.au/~hartley/Papers/fundamental/fundamental.pdf
    """
    src_h, dst_h = [_append_homogeneous_dim(pts) for pts in (src, dst)]
    Fu = F @ src_h.T
    uFu = np.sum(dst_h.T * Fu, axis=0)
    if metric == 'distance':
        # See Zhang, p 163, and Notes above.
        return uFu / np.sqrt(np.sum(Fu[:-1] ** 2, axis=0))
    if metric == 'epip-distances':
        # Hartley, p 585, section 7.1.4.
        Fu_dash = F.T @ dst_h.T
        scaler = 1 / np.sum(Fu[:-1] ** 2, axis=0) + 1 / np.sum(
            Fu_dash[:-1] ** 2, axis=0
        )
        return (uFu**2) * scaler
    raise ValueError(f'Invalid metric "{metric}"')


def test_fundamental_matrix_epipolar_projection():
    # See https://github.com/matthew-brett/test-fundamental-matrices
    # for validation of fundamental matrix estimation methods.
    src = np.array(
        [
            1.839035,
            1.924743,
            0.543582,
            0.375221,
            0.473240,
            0.142522,
            0.964910,
            0.598376,
            0.102388,
            0.140092,
            15.994343,
            9.622164,
            0.285901,
            0.430055,
            0.091150,
            0.254594,
        ]
    ).reshape(-1, 2)

    dst = np.array(
        [
            1.002114,
            1.129644,
            1.521742,
            1.846002,
            1.084332,
            0.275134,
            0.293328,
            0.588992,
            0.839509,
            0.087290,
            1.779735,
            1.116857,
            0.878616,
            0.602447,
            0.642616,
            1.028681,
        ]
    ).reshape(-1, 2)

    tform = FundamentalMatrixTransform.from_estimate(src, dst)
    assert tform

    # Calculate and test pixel distances.
    rms_ds = np.abs(_calc_distances(src, dst, tform.params))
    assert np.all(rms_ds < 0.5)

    # Check same output from utility function.
    tform_from_func = estimate_transform('fundamental', src, dst)
    assert np.allclose(tform.params, tform_from_func.params)

    # Check this corresponds to RMS scaling.
    class FMTRMS(FundamentalMatrixTransform):
        scaling = 'rms'

    tform_rms = FMTRMS.from_estimate(src, dst)
    assert tform_rms
    assert np.allclose(tform.params, tform_rms.params)

    # Check that we can also use MRS (Hartley distance).
    class FMTMRS(FundamentalMatrixTransform):
        scaling = 'mrs'

    tform_mrs = FMTMRS.from_estimate(src, dst)
    # MRS gives us a different F matrix.
    assert tform_mrs
    assert not np.allclose(tform.params, tform_mrs.params)
    # But with acceptable (in this case slightly larger) distances.
    mrs_ds = _calc_distances(src, dst, tform_mrs.params)
    assert np.all(np.abs(mrs_ds) < 0.6)

    # F matrix is as for OpenCV (CV2) (but with different scaling).
    # !pip install opencv-python-headless
    # import cv2
    # cv2.findFundamentalMat(src, dst, cv2.FM_8POINT)[0]
    cv2_f = np.array(
        [
            [-10.30598589, 19.82710842, -1.61417041],
            [-3.41654338, 2.12186874, 1.06697878],
            [11.76321795, -20.29016731, 1.0],
        ]
    )
    # OpenCV matrix scaled such that final element is 1.
    assert np.allclose(tform_mrs.params / tform_mrs.params[-1, -1], cv2_f)
    # Distances are the same (because difference is only in scaling).
    assert np.allclose(mrs_ds, _calc_distances(src, dst, cv2_f), atol=1e-7)

    # We can also (for comparison) use raw estimating (without centering or
    # scaling).
    class FMTRaw(FundamentalMatrixTransform):
        scaling = 'raw'

    tform_raw = FMTRaw.from_estimate(src, dst)
    # Raw gives us a different F matrix from RMS or MRS.
    assert tform_raw
    assert not np.allclose(tform.params, tform_raw.params)
    assert not np.allclose(tform_mrs.params, tform_raw.params)
    # Distances are greater than with either scaling option.
    raw_ds = _calc_distances(src, dst, tform_raw.params)
    assert np.max(np.abs(raw_ds)) > 1
    assert np.mean(np.abs(raw_ds)) > np.mean(np.abs(rms_ds))
    assert np.mean(np.abs(raw_ds)) > np.mean(np.abs(mrs_ds))


def test_essential_matrix_init():
    r = np.eye(3)
    t = np.array([0, 0, 1])
    tform = EssentialMatrixTransform(rotation=r, translation=t)
    assert_equal(tform.params, np.array([0, -1, 0, 1, 0, 0, 0, 0, 0]).reshape(3, 3))
    t2 = np.array([0, 0, 2])
    with pytest.raises(ValueError, match="Translation vector must have unit length"):
        EssentialMatrixTransform(rotation=r, translation=t2)
    with pytest.raises(ValueError, match="Rotation matrix must have unit determinant"):
        EssentialMatrixTransform(rotation=np.eye(3)[::-1], translation=t)
    r2 = r[[2, 0, 1]]
    tform = EssentialMatrixTransform(rotation=r2, translation=t)
    assert_equal(tform.params, [[-1, 0, 0], [0, 0, 1], [0, 0, 0]])
    with pytest.raises(ValueError):
        EssentialMatrixTransform(matrix=np.zeros((3, 2)), translation=t)
    with pytest.raises(ValueError):
        EssentialMatrixTransform(rotation=np.zeros((3, 2)), translation=t)
    with pytest.raises(ValueError):
        EssentialMatrixTransform(rotation=np.zeros((3, 3)), translation=t)
    # Both must be specified.
    with pytest.raises(ValueError):
        EssentialMatrixTransform(rotation=np.eye(3))
    # Dimensionality must match.
    with pytest.raises(ValueError):
        EssentialMatrixTransform(rotation=np.eye(3), translation=[1, 0])


def test_essential_matrix_estimation():
    src = np.array(
        [
            1.839035,
            1.924743,
            0.543582,
            0.375221,
            0.473240,
            0.142522,
            0.964910,
            0.598376,
            0.102388,
            0.140092,
            15.994343,
            9.622164,
            0.285901,
            0.430055,
            0.091150,
            0.254594,
        ]
    ).reshape(-1, 2)
    dst = np.array(
        [
            1.002114,
            1.129644,
            1.521742,
            1.846002,
            1.084332,
            0.275134,
            0.293328,
            0.588992,
            0.839509,
            0.087290,
            1.779735,
            1.116857,
            0.878616,
            0.602447,
            0.642616,
            1.028681,
        ]
    ).reshape(-1, 2)

    tform = estimate_transform('essential', src, dst)

    # Reference values obtained using COLMAP SfM library.
    tform_ref = np.array(
        [
            [-0.0811666, 0.255449, -0.0478999],
            [-0.192392, -0.0531675, 0.119547],
            [0.177784, -0.22008, -0.015203],
        ]
    )
    assert_almost_equal(tform.params, tform_ref, 6)


def test_essential_matrix_forward():
    tform = EssentialMatrixTransform(
        rotation=np.eye(3), translation=np.array([1, 0, 0])
    )
    src = np.array([[0, 0], [0, 1], [1, 1]])
    assert_almost_equal(tform(src), [[0, -1, 0], [0, -1, 1], [0, -1, 1]])


def test_essential_matrix_inverse():
    tform = EssentialMatrixTransform(
        rotation=np.eye(3), translation=np.array([1, 0, 0])
    )
    src = np.array([[0, 0], [0, 1], [1, 1]])
    assert_almost_equal(tform.inverse(src), [[0, 1, 0], [0, 1, -1], [0, 1, -1]])


def test_essential_matrix_residuals():
    tform = EssentialMatrixTransform(
        rotation=np.eye(3), translation=np.array([1, 0, 0])
    )
    src = np.array([[0, 0], [0, 0], [0, 0]])
    dst = np.array([[2, 0], [2, 1], [2, 2]])
    assert_almost_equal(tform.residuals(src, dst) ** 2, [0, 0.5, 2])


def test_projective_estimation():
    # exact solution
    tform = estimate_transform('projective', SRC[:4, :], DST[:4, :])
    assert_almost_equal(tform(SRC[:4, :]), DST[:4, :])

    # over-determined
    tform2 = estimate_transform('projective', SRC, DST)
    assert_almost_equal(tform2.inverse(tform2(SRC)), SRC)

    # via from_estimate classmethod
    tform3 = ProjectiveTransform.from_estimate(SRC, DST)
    assert_almost_equal(tform3.params, tform2.params)
    # via estimate method
    tform4 = ProjectiveTransform.identity()
    with pytest.warns(FutureWarning, match='`estimate` is deprecated'):
        assert tform4.estimate(SRC, DST)
    assert_almost_equal(tform4.params, tform2.params)


def test_projective_weighted_estimation():
    # Exact solution with same points, and unity weights
    tform = estimate_transform('projective', SRC[:4, :], DST[:4, :])
    tform_w = estimate_transform('projective', SRC[:4, :], DST[:4, :], np.ones(4))
    assert_almost_equal(tform.params, tform_w.params)

    # Over-determined solution with same points, and unity weights
    tform = estimate_transform('projective', SRC, DST)
    tform_w = estimate_transform('projective', SRC, DST, np.ones(SRC.shape[0]))
    assert_almost_equal(tform.params, tform_w.params)

    # Repeating a point, but setting its weight small, should give nearly
    # the same result.
    point_weights = np.ones(SRC.shape[0] + 1)
    point_weights[0] = 1.0e-15
    tform1 = estimate_transform('projective', SRC, DST)
    tform2 = estimate_transform(
        'projective',
        SRC[np.arange(-1, SRC.shape[0]), :],
        DST[np.arange(-1, SRC.shape[0]), :],
        point_weights,
    )
    assert_almost_equal(tform1.params, tform2.params, decimal=3)


@pytest.mark.parametrize('array_like_input', [False, True])
def test_projective_init(array_like_input):
    tform = estimate_transform('projective', SRC, DST)
    # init with transformation matrix
    if array_like_input:
        params = [list(p) for p in tform.params]
    else:
        params = tform.params
    tform2 = ProjectiveTransform(params)
    assert_almost_equal(tform2.params, tform.params)


def test_polynomial_estimation():
    # over-determined
    tform = estimate_transform('polynomial', SRC, DST, order=10)
    assert_almost_equal(tform(SRC), DST, 6)

    # via from_estimate classmethod
    tform2 = PolynomialTransform.from_estimate(SRC, DST, order=10)
    assert_almost_equal(tform2.params, tform.params)
    # via estimate method
    tform3 = PolynomialTransform()
    with pytest.warns(FutureWarning, match='`estimate` is deprecated'):
        assert tform3.estimate(SRC, DST, order=10)
    assert_almost_equal(tform3.params, tform.params)


def test_polynomial_weighted_estimation():
    # Over-determined solution with same points, and unity weights
    tform = estimate_transform('polynomial', SRC, DST, order=10)
    tform_w = estimate_transform(
        'polynomial', SRC, DST, order=10, weights=np.ones(SRC.shape[0])
    )
    assert_almost_equal(tform.params, tform_w.params)

    # Repeating a point, but setting its weight small, should give nearly
    # the same result.
    point_weights = np.ones(SRC.shape[0] + 1)
    point_weights[0] = 1.0e-15
    tform1 = estimate_transform('polynomial', SRC, DST, order=10)
    tform2 = estimate_transform(
        'polynomial',
        SRC[np.arange(-1, SRC.shape[0]), :],
        DST[np.arange(-1, SRC.shape[0]), :],
        order=10,
        weights=point_weights,
    )
    assert_almost_equal(tform1.params, tform2.params, decimal=4)


@pytest.mark.parametrize('array_like_input', [False, True])
def test_polynomial_init(array_like_input):
    tform = estimate_transform('polynomial', SRC, DST, order=10)
    # Init with transformation parameters.
    if array_like_input:
        params = [list(p) for p in tform.params]
    else:
        params = tform.params
    tform2 = PolynomialTransform(params)
    assert_almost_equal(tform2.params, tform.params)
    # Can't specify scalar params.
    with pytest.raises(ValueError):
        _ = PolynomialTransform(0)
    # Parameters must be (2, N).
    for inp in (np.eye(3), np.zeros(3)):
        with pytest.raises(ValueError):
            _ = PolynomialTransform(inp)
    # Transform always 2D.
    for d in (1, 3, 4):
        with pytest.raises(NotImplementedError):
            _ = PolynomialTransform(dimensionality=d)
        with pytest.raises(NotImplementedError):
            _ = PolynomialTransform.identity(d)


def test_polynomial_default_order():
    tform = estimate_transform('polynomial', SRC, DST)
    tform2 = estimate_transform('polynomial', SRC, DST, order=2)
    assert_almost_equal(tform2.params, tform.params)


def test_polynomial_inverse():
    with pytest.raises(NotImplementedError):
        PolynomialTransform().inverse(0)


def test_union():
    tform1 = SimilarityTransform(scale=0.1, rotation=0.3)
    tform2 = SimilarityTransform(scale=0.1, rotation=0.9)
    tform3 = SimilarityTransform(scale=0.1**2, rotation=0.3 + 0.9)
    tform = tform1 + tform2
    assert_almost_equal(tform.params, tform3.params)

    tform1 = AffineTransform(scale=(0.1, 0.1), rotation=0.3)
    tform2 = SimilarityTransform(scale=0.1, rotation=0.9)
    tform3 = SimilarityTransform(scale=0.1**2, rotation=0.3 + 0.9)
    tform = tform1 + tform2
    assert_almost_equal(tform.params, tform3.params)
    assert tform.__class__ == ProjectiveTransform

    tform = AffineTransform(scale=(0.1, 0.1), rotation=0.3)
    assert_almost_equal((tform + tform.inverse).params, np.eye(3))

    tform1 = SimilarityTransform(scale=0.1, rotation=0.3)
    tform2 = SimilarityTransform(scale=0.1, rotation=0.9)
    tform3 = SimilarityTransform(scale=0.1 * 1 / 0.1, rotation=0.3 - 0.9)
    tform = tform1 + tform2.inverse
    assert_almost_equal(tform.params, tform3.params)


def test_union_differing_types():
    tform1 = SimilarityTransform()
    tform2 = PolynomialTransform()
    with pytest.raises(TypeError):
        tform1.__add__(tform2)


@pytest.mark.parametrize(
    "tform",
    [
        ProjectiveTransform(matrix=np.random.rand(3, 3)),
        AffineTransform(scale=(0.1, 0.1), rotation=0.3),
        EuclideanTransform(rotation=0.9, translation=(5, 5)),
        SimilarityTransform(scale=0.1, rotation=0.9),
        EssentialMatrixTransform(
            rotation=np.eye(3), translation=(1 / np.sqrt(2), 1 / np.sqrt(2), 0)
        ),
        FundamentalMatrixTransform(
            matrix=EssentialMatrixTransform(
                rotation=np.eye(3), translation=(1 / np.sqrt(2), 1 / np.sqrt(2), 0)
            ).params
        ),
        PiecewiseAffineTransform.from_estimate(SRC, DST),
    ],
)
def test_inverse_all_transforms(tform):
    assert isinstance(tform.inverse, type(tform))
    if hasattr(tform, 'params'):
        assert_almost_equal(tform.inverse.inverse.params, tform.params)
    else:
        assert isinstance(tform, PiecewiseAffineTransform)
    assert_almost_equal(tform.inverse.inverse(SRC), tform(SRC))
    # Test addition with inverse, not implemented for all
    if not isinstance(
        tform,
        (
            EssentialMatrixTransform,
            FundamentalMatrixTransform,
            PiecewiseAffineTransform,
        ),
    ):
        assert_almost_equal((tform + tform.inverse)(SRC), SRC)
        assert_almost_equal((tform.inverse + tform)(SRC), SRC)


@pytest.mark.parametrize('tform_class', TRANSFORMS.values())
def test_identity(tform_class):
    if tform_class is PiecewiseAffineTransform:
        return  # Identity transform unusable.
    rng = np.random.default_rng()
    allows_nd = tform_class in HMAT_TFORMS_ND
    for ndim in (2, 3, 4, 5) if allows_nd else (2,):
        src = rng.normal(size=(10, ndim))
        t = tform_class.identity(ndim)
        if isinstance(t, FundamentalMatrixTransform):
            out = np.hstack((src, np.ones((len(src), 1))))
        else:
            out = src
        assert np.allclose(t(src), out)


def test_geometric_tform():
    with pytest.raises(TypeError, match="Can't instantiate abstract class"):
        _GeometricTransform()

    # See gh-3926 for discussion details
    for i in range(20):
        # Generate random Homography
        H = np.random.rand(3, 3) * 100
        H[2, H[2] == 0] += np.finfo(float).eps
        H /= H[2, 2]

        # Craft some src coords
        src = np.array(
            [
                [(H[2, 1] + 1) / -H[2, 0], 1],
                [1, (H[2, 0] + 1) / -H[2, 1]],
                [1, 1],
            ]
        )
        # Prior to gh-3926, under the above circumstances,
        # destination coordinates could be returned with nan/inf values.
        tform = ProjectiveTransform(H)  # Construct the transform
        dst = tform(src)  # Obtain the dst coords
        # Ensure dst coords are finite numeric values
        assert np.isfinite(dst).all()


@pytest.mark.parametrize(
    'tform_class', (FundamentalMatrixTransform, EssentialMatrixTransform)
)
def test_identical_fundamental(tform_class):
    # Test identical points to transform gives failed estimation.
    assert tform_class.from_estimate(SRC, DST)
    bad_src = np.ones((8, 2))
    bad_tform = tform_class.from_estimate(bad_src, DST)
    assert not bad_tform

    regex = "has no attribute 'params'.*Scaling failed for input points"
    with pytest.raises(FailedEstimationAccessError, match=regex):
        bad_tform.params


@pytest.mark.parametrize(
    'tform_class, msg',
    (
        (ProjectiveTransform, 'Scaling generated NaN values'),
        (AffineTransform, 'Scaling generated NaN values'),
        (EuclideanTransform, 'Poor conditioning for estimation'),
        (SimilarityTransform, 'Poor conditioning for estimation'),
    ),
)
def test_degenerate(tform_class, msg):
    src = dst = np.zeros((10, 2))

    tf = tform_class.from_estimate(src, dst)
    assert not tf
    assert str(tf) == f'{tform_class.__name__}: {msg}'
    tform = tform_class.identity()
    with pytest.warns(FutureWarning, match='`estimate` is deprecated'):
        assert not tform.estimate(src, dst)
    assert np.all(np.isnan(tform.params))


def test_degenerate_2():
    # See gh-3926 for discussion details
    tform = ProjectiveTransform()
    for i in range(20):
        # Some random coordinates
        src = np.random.rand(4, 2) * 100
        dst = np.random.rand(4, 2) * 100

        # Degenerate the case by arranging points on a single line
        src[:, 1] = np.random.rand()
        # Prior to gh-3926, under the above circumstances,
        # a transform could be returned with nan values.
        tf = ProjectiveTransform.from_estimate(src, dst)
        assert not tf
        assert str(tf) == (
            'ProjectiveTransform: Right singular vector has 0 final element'
        )
        with pytest.warns(FutureWarning, match='`estimate` is deprecated'):
            result = tform.estimate(src, dst)
        assert not result
        assert np.all(np.isnan(tform.params))

    # The tessellation on the following points produces one degenerate affine
    # warp within PiecewiseAffineTransform.
    src = np.asarray(
        [
            [0, 192, 256],
            [0, 256, 256],
            [5, 0, 192],
            [5, 64, 0],
            [5, 64, 64],
            [5, 64, 256],
            [5, 192, 192],
            [5, 256, 256],
            [0, 192, 256],
        ]
    )

    dst = np.asarray(
        [
            [0, 142, 206],
            [0, 206, 206],
            [5, -50, 142],
            [5, 14, 0],
            [5, 14, 64],
            [5, 14, 206],
            [5, 142, 142],
            [5, 206, 206],
            [0, 142, 206],
        ]
    )
    # from_estimate
    tform = PiecewiseAffineTransform.from_estimate(src, dst)
    # Simplex group index 4 has degenerate affine.
    bad_affine_i = 4
    assert not tform
    assert str(tform).startswith(
        f'PiecewiseAffineTransform: Failure at forward simplex {bad_affine_i}'
    )
    # estimate method records steps in affine estimation.
    tform = PiecewiseAffineTransform()
    with pytest.warns(FutureWarning, match='`estimate` is deprecated'):
        assert not tform.estimate(src, dst)
    # Check for degenerate affine.
    assert np.all(np.isnan(tform.affines[bad_affine_i].params))
    for idx, affine in enumerate(tform.affines):
        if idx != bad_affine_i:
            assert not np.all(np.isnan(affine.params))
    for affine in tform.inverse_affines:
        assert not np.all(np.isnan(affine.params))


def test_calc_center_normalize():
    n, d = SRC.shape
    for scaling in ('rms', 'mrs', 'raw'):
        mat = _calc_center_normalize(SRC, scaling=scaling)
        if scaling == 'raw':
            assert_equal(mat, np.eye(3))
            continue
        out_pts = _apply_homogeneous(mat, SRC)
        assert np.allclose(np.mean(out_pts, axis=0), 0)
        if scaling == 'rms':
            assert np.isclose(np.sqrt(np.mean(out_pts**2)), 1)
        elif scaling == 'mrs':
            scaler = np.mean(np.sqrt(np.sum(out_pts**2, axis=1)))
            assert np.isclose(scaler, np.sqrt(2))

            scaler = np.mean(np.sqrt(np.sum(out_pts**2, axis=1)))
            assert np.isclose(scaler, np.sqrt(2))
        mat2, normed = _center_and_normalize_points(SRC, scaling=scaling)
        assert_equal(mat, mat2)
        assert_equal(out_pts, normed)

    with pytest.raises(ValueError, match='Unexpected "scaling"'):
        _center_and_normalize_points(SRC, scaling='foo')


def test_normalize_degenerate_points():
    """Return nan matrix *of appropriate size* when point is repeated."""
    pts = np.array([[73.42834308, 94.2977623]] * 3)
    for scaling in 'rms', 'mrs':
        mat, pts_tf = _center_and_normalize_points(pts, scaling)
        assert np.all(np.isnan(mat))
        assert np.all(np.isnan(pts_tf))
        assert mat.shape == (3, 3)
        assert pts_tf.shape == pts.shape


def test_projective_repr():
    tform = ProjectiveTransform()
    want = (
        re.escape(
            textwrap.dedent(
                '''
        <ProjectiveTransform(matrix=
            [[1., 0., 0.],
             [0., 1., 0.],
             [0., 0., 1.]]) at
        '''
            ).strip()
        )
        + ' 0x[a-f0-9]+'
        + re.escape('>')
    )
    # Hack the escaped regex to allow whitespace before each number for
    # compatibility with different numpy versions.
    want = want.replace('0\\.', ' *0\\.')
    want = want.replace('1\\.', ' *1\\.')
    assert re.match(want, repr(tform))


def test_projective_str():
    tform = ProjectiveTransform()
    want = re.escape(
        textwrap.dedent(
            '''
        <ProjectiveTransform(matrix=
            [[1., 0., 0.],
             [0., 1., 0.],
             [0., 0., 1.]])>
        '''
        ).strip()
    )
    # Hack the escaped regex to allow whitespace before each number for
    # compatibility with different numpy versions.
    want = want.replace('0\\.', ' *0\\.')
    want = want.replace('1\\.', ' *1\\.')
    assert re.match(want, str(tform))


def _assert_least_squares(tf, src, dst):
    baseline = np.sum((tf(src) - dst) ** 2)
    for i in range(tf.params.size):
        for update in [0.001, -0.001]:
            params = np.copy(tf.params)
            params.flat[i] += update
            new_tf = tf.__class__(matrix=params)
            new_ssq = np.sum((new_tf(src) - dst) ** 2)
            assert new_ssq > baseline


@pytest.mark.parametrize('array_like_input', [False, True])
def test_estimate_affine_3d(array_like_input):
    ndim = 3
    src = np.random.random((25, ndim)) * 2 ** np.arange(7, 7 + ndim)
    matrix = np.array(
        [
            [4.8, 0.1, 0.2, 25],
            [0.0, 1.0, 0.1, 30],
            [0.0, 0.0, 1.0, -2],
            [0.0, 0.0, 0.0, 1.0],
        ]
    )

    if array_like_input:
        # list of lists for matrix and src coords
        src = [list(c) for c in src]
        matrix = [list(c) for c in matrix]

    tf = AffineTransform(matrix=matrix)
    dst = tf(src)
    dst_noisy = dst + np.random.random((25, ndim))
    if array_like_input:
        # list of lists for destination coords
        dst = [list(c) for c in dst]
    tf2 = AffineTransform.from_estimate(src, dst_noisy)
    # we check rot/scale/etc more tightly than translation because translation
    # estimation is on the 1 pixel scale
    matrix = np.asarray(matrix)
    assert_almost_equal(tf2.params[:, :-1], matrix[:, :-1], decimal=2)
    assert_almost_equal(tf2.params[:, -1], matrix[:, -1], decimal=0)
    _assert_least_squares(tf2, src, dst_noisy)
    tf3 = AffineTransform(dimensionality=ndim)
    with pytest.warns(FutureWarning, match='`estimate` is deprecated'):
        assert tf3.estimate(src, dst_noisy)
    assert_equal(tf2.params, tf3.params)


def test_fundamental_3d_not_implemented():
    with pytest.raises(NotImplementedError):
        _ = FundamentalMatrixTransform(dimensionality=3)
    with pytest.raises(NotImplementedError):
        _ = FundamentalMatrixTransform(np.eye(4))


def test_array_protocol():
    mat = np.eye(4)
    tf = ProjectiveTransform(mat)
    assert_equal(np.array(tf), mat)
    assert_equal(np.array(tf, dtype=int), mat.astype(int))


def test_affine_transform_from_linearized_parameters():
    mat = np.concatenate((np.random.random((3, 4)), np.eye(4)[-1:]), axis=0)
    v = mat[:-1].ravel()
    mat_from_v = _affine_matrix_from_vector(v)
    tf = AffineTransform(matrix=mat_from_v)
    assert_equal(np.array(tf), mat)
    # incorrect number of parameters
    with pytest.raises(ValueError):
        _ = _affine_matrix_from_vector(v[:-1])
    with pytest.raises(ValueError):
        _ = AffineTransform(matrix=v[:-1])


EG_OPS = dict(scale=(4, 5), shear=(1.4, 1.8), rotation=0.4, translation=(10, 12))


@pytest.mark.parametrize(
    'tform_class, op_order',
    (
        (AffineTransform, ('scale', 'shear', 'rotation', 'translation')),
        (EuclideanTransform, ('rotation', 'translation')),
        (SimilarityTransform, ('scale', 'rotation', 'translation')),
    ),
)
def test_transform_order(tform_class, op_order):
    # Test transforms are applied in order stated.
    ops = [(k, EG_OPS[k]) for k in op_order]
    part_xforms = [tform_class(**{k: v}) for k, v in ops]
    full_xform = tform_class(**dict(ops))
    # Assemble affine transform via matrices.
    out = np.eye(3)
    for tf in part_xforms:
        out = tf @ out
    assert np.allclose(full_xform.params, out)


# AffineTransform only allows 2D implicit parameters.
@pytest.mark.parametrize(
    'inp',
    (
        dict(scale=5, dimensionality=3),
        dict(scale=(5, 5, 5), dimensionality=3),
        dict(scale=(5, 5, 5)),
        dict(shear=(0.1, 0.2, 0.3)),
        dict(rotation=(0.1, 0.2)),
        dict(translation=1),
        dict(translation=(1, 2, 3)),
    ),
)
def test_affine_params_nD_error(inp):
    with pytest.raises(ValueError):
        _ = AffineTransform(**inp)


def test_euler_rotation():
    for v, angles, expected in (
        ([0, 10, 0], np.radians([90, 45, 45]), [-5, -5, 7.1]),
        ([-1, 7, -2], np.radians([-10, 23, -25]), [1.1, 6.2, -3.8]),
    ):
        R = _euler_rotation_matrix(angles)
        assert_almost_equal(R @ v, expected, decimal=1)


def _from_matvec(mat, vec):
    mat, vec = (np.array(p) for p in (mat, vec))
    d = mat.shape[0]
    out = np.eye(d + 1)
    out[:-1, :-1] = mat
    out[:-1, -1] = vec
    return out


@pytest.mark.parametrize('tform_class', (EuclideanTransform, SimilarityTransform))
def test_euclidean_param_defaults(tform_class):
    # 2D rotation is 0 when only translation is given
    tf = tform_class(translation=(5, 5))
    assert np.array(tf)[0, 1] == 0
    # off diagonals are 0 when only translation is given
    tf = tform_class(translation=(4, 5, 9), dimensionality=3)
    assert_equal(np.array(tf)[[0, 0, 1, 1, 2, 2], [1, 2, 0, 2, 0, 1]], 0)
    # Specifying translations for D>3 is supported.
    tf = tform_class(translation=(5, 6, 7, 8))
    assert_equal(tf.params, _from_matvec(np.eye(4), (5, 6, 7, 8)))
    # But must match the dimensionality.
    with pytest.raises(ValueError):
        _ = tform_class(translation=(5, 6, 7, 8), dimensionality=3)
    # Incorrect number of angles (must be 1 or 3
    with pytest.raises(ValueError):
        _ = tform_class(rotation=(4, 8))
    with pytest.raises(ValueError):
        _ = tform_class(rotation=(4, 8, 2, 4))
    # translation is 0 when rotation is given
    tf = tform_class(rotation=np.pi * np.arange(3), dimensionality=3)
    assert_equal(np.array(tf)[:-1, 3], 0)


def test_euler_angle_consistency():
    angles = np.random.random((3,)) * 2 * np.pi - np.pi
    euclid = EuclideanTransform(rotation=angles, dimensionality=3)
    similar = SimilarityTransform(rotation=angles, dimensionality=3)
    assert_array_almost_equal(euclid, similar)


def test_2D_only_implementations():
    with pytest.raises(NotImplementedError):
        _ = PolynomialTransform(dimensionality=3)
    tf = AffineTransform(dimensionality=3)
    with pytest.raises(NotImplementedError):
        _ = tf.rotation
    with pytest.raises(NotImplementedError):
        _ = tf.shear


@pytest.mark.parametrize('tform_class', HMAT_TFORMS)
def test_kw_only_params(tform_class):
    # Check only matrix can be passed as positional arg.
    with pytest.raises(TypeError):
        tform_class(None, None)


def test_kw_only_emt():
    # Check all parameters are keyword-only for EssentialMatrixTransform.
    with pytest.raises(TypeError):
        EssentialMatrixTransform(None)


@pytest.mark.parametrize('tform_class', HMAT_TFORMS)
def test_init_contract_dims(tform_class):
    allows_nd = tform_class in HMAT_TFORMS_ND
    # 2D identity is default.
    for tf in (
        tform_class(),
        tform_class(dimensionality=2),
        tform_class.identity(),
        tform_class.identity(None),
        tform_class(None),
    ):
        assert_equal(tf.params, np.eye(3))
    ok_dims = (2, 3, 4, 5) if allows_nd else (2,)
    for d in ok_dims:
        h_d = d + 1
        # Identity for acceptable dimensions.
        for tf in (
            tform_class(dimensionality=d),
            tform_class.identity(d),
            tform_class.identity(dimensionality=d),
        ):
            assert_equal(tf.params, np.eye(h_d))
        # Wrong shape for given dimensions.
        for matrix in np.eye(h_d)[:-1], np.eye(h_d)[:, :-1]:
            with pytest.raises(ValueError):
                tform_class(matrix)
    err_dims = (1,) if allows_nd else (1, 3, 4, 5)
    for matrix in [np.eye(d + 1) for d in err_dims]:
        with pytest.raises(NotImplementedError):
            tform_class(matrix)
    # Test vector matrix input invalid.
    with pytest.raises(ValueError):
        tform_class(np.zeros((2, 3)))


def test_astronaut_piecewise():
    # From doc/examples/transforms/plot_piecewise_affine.py
    image = data.astronaut()
    rows, cols = image.shape[0], image.shape[1]

    src_cols = np.linspace(0, cols, 20)
    src_rows = np.linspace(0, rows, 10)
    src_rows, src_cols = np.meshgrid(src_rows, src_cols)
    src = np.dstack([src_cols.flat, src_rows.flat])[0]

    # add sinusoidal oscillation to row coordinates
    dst_rows = src[:, 1] - np.sin(np.linspace(0, 3 * np.pi, src.shape[0])) * 50
    dst_cols = src[:, 0]
    dst_rows *= 1.5
    dst_rows -= 1.5 * 50
    dst = np.vstack([dst_cols, dst_rows]).T

    # Transform will fail with strict check for rank deficiency of
    # inverse matrices in ProjectiveTransform.estimate.
    assert PiecewiseAffineTransform.from_estimate(src, dst)


def test_broadcasting():
    # Scalar scale broadcasts.
    translation = [3, 4, 5]
    tf = SimilarityTransform(scale=2, translation=translation)
    assert_equal(tf.params, _from_matvec(np.eye(3) * 2, translation))
    # Translation does broadcast.
    # 2D.
    tf = SimilarityTransform(scale=2, translation=10)
    assert_equal(tf.params, _from_matvec(np.eye(2) * 2, [10, 10]))
    # 3D.
    tf = SimilarityTransform(scale=[2, 3, 4], translation=10)
    assert_equal(tf.params, _from_matvec(np.diag([2, 3, 4]), [10] * 3))
    # Scalar rotation does not broadcast.
    for tf_class in SimilarityTransform, EuclideanTransform:
        with pytest.raises(ValueError):
            tf_class(rotation=0.2, translation=translation)


@pytest.mark.parametrize('tf_class', TRANSFORMS.values())
def test_estimate_futurewarn(tf_class):
    tf = tf_class.identity()
    msg = (
        f'`estimate` is deprecated since .* Please use `{tf_class.__name__}'
        '.from_estimate` class constructor instead.'
    )
    with pytest.warns(FutureWarning, match=msg):
        assert tf.estimate(SRC, DST)



#!/usr/bin/env python
"""Generate requirements/*.txt files from pyproject.toml.

Also builda a conda environment.yml
"""

import sys
import re
from pathlib import Path

try:  # standard module since Python 3.11
    import tomllib as toml
except ImportError:
    try:  # available for older Python via pip
        import tomli as toml
    except ImportError:
        sys.exit("Please install `tomli` first: `pip install tomli`")

script_pth = Path(__file__)
repo_dir = script_pth.parent.parent
script_relpth = script_pth.relative_to(repo_dir)
header = [
    f"# Generated via {script_relpth.as_posix()} and pre-commit hook.",
    "# Do not edit this file; modify pyproject.toml instead.",
]


def generate_requirement_file(name: str, req_list: list[str]) -> None:
    req_fname = repo_dir / "requirements" / f"{name}.txt"
    req_fname.write_text("\n".join(header + req_list) + "\n")


def generate_environment_yml(req_sections: dict[str, list[str]]) -> None:
    # Some packages in conda-forge have different names than they do on pypi
    # Also, some packages have a 'base' flavoured conda package which
    # doesn't install all their optional dependencies.
    rename_idx = {
        'build': 'python-build',
        'kaleido': 'python-kaleido',
        'sphinx_design': 'sphinx-design',
        'astropy': 'astropy-base',
        'matplotlib': 'matplotlib-base',
    }
    lines = ["name: skimage-dev", "channels:", "  - conda-forge", "dependencies:"]
    for section in req_sections:
        lines.append(f"  # {section}")
        for dep in req_sections[section]:
            # Remove optional specifiers such as `[parallel]`
            dep = re.sub('\\[.*?\\]', '', dep)

            # Remove platform specifiers such as `; sys_platform != "emscripten"`
            dep = re.sub('; .*', '', dep)

            pkgname = re.split('[>=]', dep)[0]
            dep = dep.replace(pkgname, rename_idx.get(pkgname, pkgname))
            if dep == "scikit-image":
                continue

            lines.append(f"  - {dep}")

    with open("environment.yml", "w") as f:
        f.writelines(f"{line}\n" for line in lines)


def expand_dependencies(
    dep_list: list[str],
    optional_dict: dict[str, list[str]],
    package_name: str = 'scikit-image',
) -> list[str]:
    """Explode dependencies with optional extras into a flat list.

    If `scikit-image[optional_group]` is used as a dependency itself, replace
    with the actual dependencies of `optional_group`.
    """
    exploded = []
    for dep in dep_list:
        if dep.startswith(package_name):
            extras = dep.split('[')[1].rstrip(']').split(',')
            exploded.extend(
                expand_dependencies(
                    optional_dict[extras[0]], optional_dict, package_name
                )
            )
        else:
            exploded.append(dep)
    return exploded


def main() -> None:
    pyproject = toml.loads((repo_dir / "pyproject.toml").read_text())

    generate_requirement_file("default", pyproject["project"]["dependencies"])

    for key, opt_list in pyproject["project"]["optional-dependencies"].items():
        generate_requirement_file(
            key,
            expand_dependencies(
                opt_list,
                pyproject["project"]["optional-dependencies"],
                package_name='scikit-image',
            ),
        )

    generate_environment_yml(
        {
            **{'base': pyproject["project"]["dependencies"]},
            **pyproject["project"]["optional-dependencies"],
        }
    )


if __name__ == "__main__":
    main()



import importlib

import pytest

import skimage.metrics
from skimage.util import _backends


def mock_public_api_module(func):
    # Unlike the real `public_api_module` this returns the actual
    # full module name and does not perform a sanity check
    # because that test would fail for the functions we define
    # inside our tests.
    return func.__module__


@pytest.fixture
def fake_backends(monkeypatch):
    """Mock backend setup

    Two backends, with one function each.
    """

    class Backend1:
        def get_implementation(self, name):
            def fake_foo(x):
                return x * 3

            if not name.endswith(":foo"):
                raise ValueError(
                    "Backend only implements the 'foo' function."
                    f" Called with '{name}'"
                )

            return fake_foo

        def can_has(self, name, *args, **kwargs):
            if not name.endswith(":foo"):
                raise ValueError(
                    "Backend only implements the 'foo' function."
                    f" Called with '{name}'"
                )
            return True

    class Backend2:
        def get_implementation(self, name):
            def fake_foo(x):
                return x * 4

            if not name.endswith(":foo"):
                return None

            return fake_foo

        def can_has(self, name, *args, **kwargs):
            if name.endswith(":foo"):
                return True
            else:
                return False

    class BackendEntryPoint1:
        def load(self):
            return Backend1()

    class BackendEntryPoint2:
        def load(self):
            return Backend2()

    def mock_all_backends():
        return {
            "fake2": {
                "implementation": BackendEntryPoint2(),
                "info": _backends.BackendInformation(["test_backends:foo"]),
            },
            "fake1": {
                "implementation": BackendEntryPoint1(),
                "info": _backends.BackendInformation(["test_backends:foo"]),
            },
        }

    monkeypatch.setattr(_backends, "all_backends", mock_all_backends)
    monkeypatch.setattr(_backends, "public_api_module", mock_public_api_module)


@pytest.fixture
def no_backends(monkeypatch):
    """Mock backend setup with no backends"""

    def mock_no_backends():
        return {}

    monkeypatch.setattr(_backends, "all_backends", mock_no_backends)
    monkeypatch.setattr(_backends, "public_api_module", mock_public_api_module)


def test_no_notification_without_backends(no_backends):
    # Check that no DispatchNotification is raised when no backend
    # is installed.
    @_backends.dispatchable
    def foo(x):
        return x * 2

    r = foo(42)

    assert r == 42 * 2


def test_no_dispatching_when_disabled(fake_backends, monkeypatch):
    monkeypatch.setenv("SKIMAGE_NO_DISPATCHING", "1")

    @_backends.dispatchable
    def foo(x):
        return x * 2

    r = foo(42)

    assert r == 42 * 2


def test_notification_raised(fake_backends):
    @_backends.dispatchable
    def foo(x):
        return x * 2

    with pytest.warns(
        _backends.DispatchNotification,
        # Checking for fake1 means we also check that the backends are
        # used in the correct priority/order
        match="Call to.*:foo' was dispatched to the 'fake1' backend",
    ):
        r = foo(42)

    assert r == 42 * 3


@pytest.mark.parametrize(
    "func, expected",
    [
        (skimage.metrics.mean_squared_error, "skimage.metrics"),
        (skimage.io.concatenate_images, "skimage.io"),
        (skimage.filters.rank.entropy, "skimage.filters.rank"),
    ],
)
def test_module_name_determination(func, expected):
    module_name = _backends.public_api_module(func)

    assert module_name == expected

    mod = importlib.import_module(module_name)
    assert getattr(mod, func.__name__) is func



#!/usr/bin/env python3

"""Determine and print version number.

Used in top level ``meson.build``.
"""

import subprocess
from pathlib import Path


def version_from_init():
    """Extract version string from ``skimage/__init__.py``."""
    skimage_init = Path(__file__).parent / '../__init__.py'
    assert skimage_init.is_file()

    with skimage_init.open("r") as file:
        data = file.readlines()

    version_line = next(line for line in data if line.startswith('__version__ ='))
    version = version_line.strip().split(' = ')[1].replace('"', '').replace("'", '')
    return version


def append_git_revision_and_date(version):
    """Try to append last commit date and hash to version.

    Appends nothing if the current working directory is outside a git
    repository.
    """
    try:
        result = subprocess.run(
            ['git', 'log', '-1', '--format="%H %aI"'],
            capture_output=True,
            text=True,
            check=True,
        )
    except subprocess.CalledProcessError:
        pass
    else:
        git_hash, git_date = (
            result.stdout.strip()
            .replace('"', '')
            .split('T')[0]
            .replace('-', '')
            .split()
        )
        version += f'+git{git_date}.{git_hash[:7]}'
    return version


if __name__ == "__main__":
    version = version_from_init()
    if 'dev' in version:
        version = append_git_revision_and_date(version)
    print(version)



"""
==============================================================
Use rolling-ball algorithm for estimating background intensity
==============================================================

The rolling-ball algorithm estimates the background intensity of a grayscale
image. It comes in useful, for instance, in case of uneven exposure, when
subtracting the background is desirable. It is frequently used in biomedical
image processing and was first proposed by Stanley R. Sternberg in
1983 [1]_.

The algorithm works as a filter: Think of the image
as a surface that has unit-sized blocks stacked on top of each other in place
of each pixel. The number of blocks, and hence surface height, is determined
by the intensity of the pixel. To get the intensity of the background at a
desired (pixel) position, we imagine submerging a ball under the surface at the
desired position. Once it is completely covered by the blocks, the apex of
the ball determines the intensity of the background at that position. We can
then 'roll' this ball around below the surface to get the background values for
the entire image. The larger the ball, the smoother the background.

scikit-image implements a generalized version of this rolling-ball algorithm,
allowing you to work with n-dimensional images and to use not only balls, but
other kernels as well. This way, you may directly filter RGB images or
image stacks along any (or all) spatial dimensions.

.. [1] Sternberg, Stanley R. "Biomedical image processing." Computer 1 (1983):
    22-34. :DOI:`10.1109/MC.1983.1654163`


Classic rolling ball
--------------------

In scikit-image, the implementation assumes that your image background has
low intensity (dark), whereas the features have high intensity (bright). If
this is not your case, you first have to invert the image—we give an example
of that further on.

"""

import matplotlib.pyplot as plt
import numpy as np

import pywt
import skimage as ski


def plot_result(image, background):
    fig, ax = plt.subplots(ncols=3)

    ax[0].imshow(image, cmap='gray')
    ax[0].set_title('Original image')
    ax[0].axis('off')

    ax[1].imshow(background, cmap='gray')
    ax[1].set_title('Background')
    ax[1].axis('off')

    ax[2].imshow(image - background, cmap='gray')
    ax[2].set_title('Result')
    ax[2].axis('off')

    fig.tight_layout()


image = ski.data.coins()

background = ski.restoration.rolling_ball(image)

plot_result(image, background)
plt.show()

######################################################################
# Bright background
# -----------------
#
# If you have dark features on a bright background, you need to invert
# the image before passing it to the rolling-ball filter, and then invert the
# result. This can be accomplished as follows:

image = ski.data.page()
image_inverted = ski.util.invert(image)

background_inverted = ski.restoration.rolling_ball(image_inverted, radius=45)
filtered_image_inverted = image_inverted - background_inverted
filtered_image = ski.util.invert(filtered_image_inverted)
background = ski.util.invert(background_inverted)

fig, ax = plt.subplots(ncols=3)

ax[0].imshow(image, cmap='gray')
ax[0].set_title('Original image')
ax[0].axis('off')

ax[1].imshow(background, cmap='gray')
ax[1].set_title('Background')
ax[1].axis('off')

ax[2].imshow(filtered_image, cmap='gray')
ax[2].set_title('Result')
ax[2].axis('off')

fig.tight_layout()

plt.show()

######################################################################
# Be careful not to fall victim to an integer underflow when subtracting
# a bright background. For example, this code looks correct, but may
# suffer from an underflow leading to unwanted artifacts. You can see
# this in the top right corner of the visualization.

image = ski.data.page()
image_inverted = ski.util.invert(image)

background_inverted = ski.restoration.rolling_ball(image_inverted, radius=45)
background = ski.util.invert(background_inverted)
underflow_image = image - background  # integer underflow occurs here

# correct subtraction
correct_image = ski.util.invert(image_inverted - background_inverted)

fig, ax = plt.subplots(ncols=2)

ax[0].imshow(underflow_image, cmap='gray')
ax[0].set_title('Background Removal with Underflow')
ax[0].axis('off')

ax[1].imshow(correct_image, cmap='gray')
ax[1].set_title('Correct Background Removal')
ax[1].axis('off')

fig.tight_layout()

plt.show()

######################################################################
# Image Datatypes
# ---------------
#
# ``rolling_ball`` can handle datatypes other than `np.uint8`. You can
# pass them into the function in the same way.

image = ski.data.coins()[:200, :200].astype(np.uint16)

background = ski.restoration.rolling_ball(image, radius=70.5)
plot_result(image, background)
plt.show()

######################################################################
# However, you need to be careful if you use floating point images
# that have been normalized to ``[0, 1]``. In this case the ball will
# be much larger than the image intensity, which can lead to
# unexpected results.

image = ski.util.img_as_float(ski.data.coins()[:200, :200])

background = ski.restoration.rolling_ball(image, radius=70.5)
plot_result(image, background)
plt.show()

######################################################################
# Because ``radius=70.5`` is much larger than the maximum intensity of
# the image, the effective kernel size is reduced significantly, i.e.,
# only a small cap (approximately ``radius=10``) of the ball is rolled
# around in the image. You can find a reproduction of this strange
# effect in the ``Advanced Shapes`` section below.
#
# To get the expected result, you need to reduce the intensity of the
# kernel. This is done by specifying the kernel manually using the
# ``kernel`` argument.
#
# Note: The radius is equal to the length of a semi-axis of an
# ellipsis, which is *half* a full axis. Hence, the kernel shape is
# multiplied by two.

normalized_radius = 70.5 / 255
image = ski.util.img_as_float(ski.data.coins())
kernel = ski.restoration.ellipsoid_kernel((70.5 * 2, 70.5 * 2), normalized_radius * 2)

background = ski.restoration.rolling_ball(image, kernel=kernel)
plot_result(image, background)
plt.show()

######################################################################
# Advanced Shapes
# ---------------
#
# By default, :func:`skimage.restoration.rolling_ball` uses a ball-shaped
# kernel (surprise).
# Sometimes, though, this can be too limiting---as in the example above,
# because the intensity dimension has a different scale compared to
# the spatial dimensions, or because the image dimensions may have
# different meanings (e.g., one could be a stack counter in an image stack).
#
# To account for this, :func:`skimage.restoration.rolling_ball` has a `kernel`
# argument which allows you to specify the kernel to be used. A kernel must
# have the same dimensionality as the image (i.e., the same number of
# dimensions/axes).
# To help with its creation, two kernel implementations are provided:
# :func:`skimage.restoration.ball_kernel` specifies a ball-shaped kernel and
# is used as the default kernel; :func:`skimage.restoration.ellipsoid_kernel`
# specifies an ellipsoid-shaped kernel.

image = ski.data.coins()
kernel = ski.restoration.ellipsoid_kernel((70.5 * 2, 70.5 * 2), 70.5 * 2)

background = ski.restoration.rolling_ball(image, kernel=kernel)
plot_result(image, background)
plt.show()

######################################################################
# You can also use :func:`skimage.restoration.ellipsoid_kernel` to recreate
# the previous, unexpected result and see that the effective (spatial) filter
# size was reduced.

image = ski.data.coins()
kernel = ski.restoration.ellipsoid_kernel((10 * 2, 10 * 2), 255 * 2)

background = ski.restoration.rolling_ball(image, kernel=kernel)
plot_result(image, background)
plt.show()

######################################################################
# Higher Dimensions
# -----------------
#
# Another interesting feature of :func:`skimage.restoration.rolling_ball` is
# that you can directly apply it to higher-dimensional images, e.g., a Z-stack
# of images obtained in confocal microscopy. The number of kernel dimensions
# must match that of image dimensions; in this case, the kernel is thus
# 3-dimensional.

image = ski.data.cells3d()[:, 1, ...]
kernel = ski.restoration.ellipsoid_kernel((1, 21, 21), 0.1)

background = ski.restoration.rolling_ball(image, kernel=kernel)

plot_result(image[30, ...], background[30, ...])
plt.show()

#####################################################################
# The above filter is actually applied to each image in the stack
# *individually*, since the kernel has size 1 along the Z axis.
#
# To filter along all 3 dimensions at the same time, you must use sizes
# greater than 1 along all 3 dimensions.
#
#   .. code-block:: python
#      :caption: Change the first line in the above code block.
#
#       kernel = ski.restoration.ellipsoid_kernel((5, 21, 21), 0.1)

######################################################################
# Another possibility is to filter individual pixels *only* along the
# planar (Z) axis.
#
#   .. code-block:: python
#      :caption: This change here will show a lot more clearly in the result.
#
#       kernel = ski.restoration.ellipsoid_kernel((100, 1, 1), 0.1)

######################################################################
# 1D Signal Filtering
# -------------------
#
# As another example of the n-dimensional feature of
# ``rolling_ball``, we show an implementation for 1D data. Here,
# we are interested in removing the background signal of an ECG waveform
# to detect prominent peaks (higher values than the local baseline).
# Smoother peaks can be removed with smaller values of the radius.

x = pywt.data.ecg()
background = ski.restoration.rolling_ball(x, radius=80)
background2 = ski.restoration.rolling_ball(x, radius=10)
plt.figure()
plt.plot(x, label='original')
plt.plot(x - background, label='radius=80')
plt.plot(x - background2, label='radius=10')
plt.legend()
plt.show()



# cython: cdivision=True
# cython: boundscheck=False
# cython: nonecheck=False
# cython: wraparound=False

import numpy as np
cimport numpy as cnp

from cpython.mem cimport PyMem_Malloc, PyMem_Free
from libc.stdlib cimport labs
from libc.math cimport fabs, sqrt, ceil, atan2, M_PI

from ..draw import circle_perimeter

from .._shared.interpolation cimport round

cnp.import_array()


def _hough_circle(cnp.ndarray img,
                  cnp.ndarray[ndim=1, dtype=cnp.intp_t] radius,
                  char normalize=True, char full_output=False):
    """Perform a circular Hough transform.

    Parameters
    ----------
    img : (M, N) ndarray
        Input image with nonzero values representing edges.
    radius : ndarray
        Radii at which to compute the Hough transform.
    normalize : boolean, optional (default True)
        Normalize the accumulator with the number
        of pixels used to draw the radius.
    full_output : boolean, optional (default False)
        Extend the output size by twice the largest
        radius in order to detect centers outside the
        input picture.

    Returns
    -------
    H : ndarray, shape (radius index, M + 2R, N + 2R)
        Hough transform accumulator for each radius.
        R designates the larger radius if full_output is True.
        Otherwise, R = 0.
    """
    if img.ndim != 2:
        raise ValueError('The input image must be 2D.')

    cdef Py_ssize_t xmax = img.shape[0]
    cdef Py_ssize_t ymax = img.shape[1]

    # compute the nonzero indexes
    cdef cnp.ndarray[ndim=1, dtype=cnp.intp_t] x, y
    x, y = np.nonzero(img)

    cdef Py_ssize_t num_pixels = x.size

    cdef Py_ssize_t offset = 0
    if full_output:
        # Offset the image
        offset = radius.max()
        x = x + offset
        y = y + offset

    cdef Py_ssize_t i, p, c, num_circle_pixels, tx, ty
    cdef cnp.float64_t incr
    cdef cnp.ndarray[ndim=1, dtype=cnp.intp_t] circle_x, circle_y

    cdef cnp.ndarray[ndim=3, dtype=cnp.float64_t] acc = \
         np.zeros((radius.size,
                   img.shape[0] + 2 * offset,
                   img.shape[1] + 2 * offset), dtype=np.float64)

    for i, rad in enumerate(radius):
        # Store in memory the circle of given radius
        # centered at (0,0)
        circle_x, circle_y = circle_perimeter(0, 0, rad)

        num_circle_pixels = circle_x.size

        with nogil:

            if normalize:
                incr = 1.0 / num_circle_pixels
            else:
                incr = 1

            # For each non zero pixel
            for p in range(num_pixels):
                # Plug the circle at (px, py),
                # its coordinates are (tx, ty)
                for c in range(num_circle_pixels):
                    tx = circle_x[c] + x[p]
                    ty = circle_y[c] + y[p]
                    if offset:
                        acc[i, tx, ty] += incr
                    elif 0 <= tx < xmax and 0 <= ty < ymax:
                        acc[i, tx, ty] += incr

    return acc


def _hough_ellipse(cnp.ndarray img, Py_ssize_t threshold=4,
                   cnp.float64_t accuracy=1, Py_ssize_t min_size=4,
                   max_size=None):
    """Perform an elliptical Hough transform.

    Parameters
    ----------
    img : (M, N) ndarray
        Input image with nonzero values representing edges.
    threshold: int, optional (default 4)
        Accumulator threshold value.
    accuracy : float64, optional (default 1)
        Bin size on the minor axis used in the accumulator.
    min_size : int, optional (default 4)
        Minimal major axis length.
    max_size : int, optional
        Maximal minor axis length. (default None)
        If None, the value is set to the half of the smaller
        image dimension.

    Returns
    -------
    result : ndarray with fields [(accumulator, yc, xc, a, b, orientation)]
          Where ``(yc, xc)`` is the center, ``(a, b)`` the major and minor
          axes, respectively. The `orientation` value follows
          `skimage.draw.ellipse_perimeter` convention.

    Examples
    --------
    >>> img = np.zeros((25, 25), dtype=np.uint8)
    >>> rr, cc = ellipse_perimeter(10, 10, 6, 8)
    >>> img[cc, rr] = 1
    >>> result = hough_ellipse(img, threshold=8)
    [(10, 10.0, 8.0, 6.0, 0.0, 10.0)]

    Notes
    -----
    The accuracy must be chosen to produce a peak in the accumulator
    distribution. In other words, a flat accumulator distribution with low
    values may be caused by a too low bin size.

    References
    ----------
    .. [1] Xie, Yonghong, and Qiang Ji. "A new efficient ellipse detection
           method." Pattern Recognition, 2002. Proceedings. 16th International
           Conference on. Vol. 2. IEEE, 2002
    """
    if img.ndim != 2:
            raise ValueError('The input image must be 2D.')

    # The creation of the array `pixels` results in a rather nasty error
    # when the image is empty.
    # As discussed in GitHub #2820 and #2996, we opt to return an empty array.
    if not np.any(img):
        return np.zeros((0, 6))

    cdef Py_ssize_t[:, ::1] pixels = np.vstack(np.nonzero(img))

    cdef Py_ssize_t num_pixels = pixels.shape[1]
    cdef list acc = list()
    cdef list results = list()
    cdef cnp.float64_t bin_size = accuracy * accuracy

    cdef cnp.float64_t max_b_squared
    if max_size is None:
        if img.shape[0] < img.shape[1]:
            max_b_squared = np.round(0.5 * img.shape[0])
        else:
            max_b_squared = np.round(0.5 * img.shape[1])
        max_b_squared *= max_b_squared
    else:
        max_b_squared = max_size * max_size

    cdef Py_ssize_t p1, p2, p3, p1x, p1y, p2x, p2y, p3x, p3y
    cdef cnp.float64_t xc, yc, a, b, d, k, dx, dy
    cdef cnp.float64_t cos_tau_squared, b_squared, orientation

    for p1 in range(num_pixels):
        p1x = pixels[1, p1]
        p1y = pixels[0, p1]

        for p2 in range(p1):
            p2x = pixels[1, p2]
            p2y = pixels[0, p2]

            # Candidate: center (xc, yc) and main axis a
            dx = p1x - p2x
            dy = p1y - p2y
            a = 0.5 * sqrt(dx * dx + dy * dy)
            if a > 0.5 * min_size:
                xc = 0.5 * (p1x + p2x)
                yc = 0.5 * (p1y + p2y)

                for p3 in range(num_pixels):
                    p3x = pixels[1, p3]
                    p3y = pixels[0, p3]
                    dx = p3x - xc
                    dy = p3y - yc
                    d = sqrt(dx * dx + dy * dy)
                    if d > min_size:
                        dx = p3x - p1x
                        dy = p3y - p1y
                        cos_tau_squared = ((a*a + d*d - dx*dx - dy*dy)
                                           / (2 * a * d))
                        cos_tau_squared *= cos_tau_squared
                        # Consider b2 > 0 and avoid division by zero
                        k = a*a - d*d * cos_tau_squared
                        if k > 0 and cos_tau_squared < 1:
                            b_squared = a*a * d*d * (1 - cos_tau_squared) / k
                            # b2 range is limited to avoid histogram memory
                            # overflow
                            if b_squared <= max_b_squared:
                                acc.append(b_squared)

                if len(acc) > 0:
                    bins = np.arange(0, np.max(acc) + bin_size, bin_size)
                    hist, bin_edges = np.histogram(acc, bins=bins)
                    hist_max = np.max(hist)
                    if hist_max > threshold:
                        orientation = atan2(p1x - p2x, p1y - p2y)
                        b = sqrt(bin_edges[hist.argmax()])
                        # to keep ellipse_perimeter() convention
                        if orientation != 0:
                            orientation = M_PI - orientation
                            # When orientation is not in [-pi:pi]
                            # it would mean in ellipse_perimeter()
                            # that a < b. But we keep a > b.
                            if orientation > M_PI:
                                orientation = orientation - M_PI / 2.
                                a, b = b, a
                        results.append((hist_max,  # Accumulator
                                        yc, xc,
                                        a, b,
                                        orientation))
                    acc = []

    return np.array(results, dtype=[('accumulator', np.intp),
                                    ('yc', np.float64),
                                    ('xc', np.float64),
                                    ('a', np.float64),
                                    ('b', np.float64),
                                    ('orientation', np.float64)])


def _hough_line(cnp.ndarray img,
                cnp.ndarray[ndim=1, dtype=cnp.float64_t] theta):
    """Perform a straight line Hough transform.

    Parameters
    ----------
    img : (M, N) ndarray
        Input image with nonzero values representing edges.
    theta : 1D ndarray of float64
        Angles at which to compute the transform, in radians.

    Returns
    -------
    H : (P, Q) ndarray of uint64
        Hough transform accumulator.
    theta : ndarray
        Angles at which the transform was computed, in radians.
    distances : ndarray
        Distance values.

    Notes
    -----
    The origin is the top left corner of the original image.
    X and Y axis are horizontal and vertical edges respectively.
    The distance is the minimal algebraic distance from the origin
    to the detected line.

    Examples
    --------
    Generate a test image:

    >>> img = np.zeros((100, 150), dtype=bool)
    >>> img[30, :] = 1
    >>> img[:, 65] = 1
    >>> img[35:45, 35:50] = 1
    >>> for i in range(90):
    ...     img[i, i] = 1
    >>> rng = np.random.default_rng()
    >>> img += rng.random(img.shape) > 0.95

    Apply the Hough transform:

    >>> out, angles, d = hough_line(img)

    .. plot:: hough_tf.py

    """
    # Compute the array of angles and their sine and cosine
    cdef cnp.ndarray[ndim=1, dtype=cnp.float64_t] ctheta
    cdef cnp.ndarray[ndim=1, dtype=cnp.float64_t] stheta

    ctheta = np.cos(theta)
    stheta = np.sin(theta)

    # compute the bins and allocate the accumulator array
    cdef cnp.ndarray[ndim=2, dtype=cnp.uint64_t] accum
    cdef cnp.ndarray[ndim=1, dtype=cnp.float64_t] bins
    cdef Py_ssize_t max_distance, offset

    offset = <Py_ssize_t>ceil(sqrt(img.shape[0] * img.shape[0] +
                                   img.shape[1] * img.shape[1]))
    max_distance = 2 * offset + 1
    accum = np.zeros((max_distance, theta.shape[0]), dtype=np.uint64)
    bins = np.linspace(-offset, offset, max_distance)

    # compute the nonzero indexes
    cdef cnp.ndarray[ndim=1, dtype=cnp.npy_intp] x_idxs, y_idxs
    y_idxs, x_idxs = np.nonzero(img)

    # finally, run the transform
    cdef Py_ssize_t nidxs, nthetas, i, j, x, y, accum_idx

    nidxs = y_idxs.shape[0]  # x and y are the same shape
    nthetas = theta.shape[0]
    with nogil:
        for i in range(nidxs):
            x = x_idxs[i]
            y = y_idxs[i]
            for j in range(nthetas):
                accum_idx = round((ctheta[j] * x + stheta[j] * y)) + offset
                accum[accum_idx, j] += 1

    return accum, theta, bins


def _probabilistic_hough_line(cnp.ndarray img, Py_ssize_t threshold,
                              Py_ssize_t line_length, Py_ssize_t line_gap,
                              cnp.ndarray[ndim=1, dtype=cnp.float64_t] theta,
                              rng=None):
    """Return lines from a progressive probabilistic line Hough transform.

    Parameters
    ----------
    img : (M, N) ndarray
        Input image with nonzero values representing edges.
    threshold : int
        Threshold in the accumulator to detect lines against noise.
    line_length : int
        Minimum accepted length of detected lines.
        Increase the parameter to extract longer lines.
    line_gap : int
        Maximum gap between pixels to still form a line.
        Increase the parameter to merge broken lines more aggressively.
    theta : (K,) ndarray of float64
        Angles at which to compute the transform, in radians.
    rng : {`numpy.random.Generator`, int}, optional
        Pseudo-random number generator.
        By default, a PCG64 generator is used (see :func:`numpy.random.default_rng`).
        If `rng` is an int, it is used to seed the generator.

    Returns
    -------
    lines : list
        List of lines identified, lines in format ((x0, y0), (x1, y1)),
        indicating line start and end.

    References
    ----------
    .. [1] C. Galamhos, J. Matas and J. Kittler, "Progressive probabilistic
           Hough transform for line detection", in IEEE Computer Society
           Conference on Computer Vision and Pattern Recognition, 1999.

    Notes
    -----

    The algorithm (from [1]_) is the following:

    1. Check the input image, if it is empty then finish.
    2. Update the accumulator with a single pixel randomly selected from the
       input image.
    3. Remove pixel from input image.
    4. Check if the highest peak in the accumulator that was modified by the
       new pixel is higher than threshold. If not then goto 1.
    5. Look along a corridor specified by the peak in the accumulator, and find
       the longest segment of pixels either continuous or exhibiting a gap not
       exceeding a given threshold.
    6. Remove the pixels in the segment from input image.
    7. Unvote from the accumulator all the pixels from the line that have
       previously voted.
    8. If the line segment is longer than the minimum length add it into the
       output list.
    9. goto 1.

    """
    cdef Py_ssize_t height = img.shape[0]
    cdef Py_ssize_t width = img.shape[1]

    # compute the bins and allocate the accumulator array
    cdef cnp.ndarray[ndim=2, dtype=cnp.uint8_t] mask = \
        np.zeros((height, width), dtype=np.uint8)
    cdef Py_ssize_t *line_end = \
        <Py_ssize_t *>PyMem_Malloc(4 * sizeof(Py_ssize_t))
    if not line_end:
        raise MemoryError('could not allocate line_end')
    cdef Py_ssize_t max_distance, offset, index
    cdef cnp.float64_t line_sin, line_cos, a, b, rho
    cdef Py_ssize_t j, k, x, y, px, py, accum_idx, max_theta_idx, rho_idx
    cdef Py_ssize_t xflag, x0, y0, dx0, dy0, dx, dy, gap, x1, y1, count
    cdef cnp.int64_t value, max_value,
    cdef int shift = 16
    cdef int good_line
    cdef Py_ssize_t nlines = 0
    cdef Py_ssize_t lines_max = 2 ** 15  # maximum line number cutoff
    cdef cnp.intp_t[:, :, ::1] lines = np.zeros((lines_max, 2, 2),
                                                dtype=np.intp)
    max_distance = 2 * <Py_ssize_t>ceil((sqrt(img.shape[0] * img.shape[0] +
                                              img.shape[1] * img.shape[1])))
    cdef cnp.int64_t[:, ::1] accum = np.zeros((max_distance, theta.shape[0]),
                                              dtype=np.int64)
    offset = max_distance / 2
    cdef Py_ssize_t nthetas = theta.shape[0]

    # compute sine and cosine of angles
    cdef cnp.float64_t[::1] ctheta = np.cos(theta)
    cdef cnp.float64_t[::1] stheta = np.sin(theta)

    # find the nonzero indexes
    cdef cnp.intp_t[:] y_idxs, x_idxs
    y_idxs, x_idxs = np.nonzero(img)

    # mask all non-zero indexes
    mask[y_idxs, x_idxs] = 1

    count = len(x_idxs)
    random_state = np.random.default_rng(rng)
    random_ = np.arange(count, dtype=np.intp)
    random_state.shuffle(random_)
    cdef cnp.intp_t[::1] random = random_

    with nogil:
        while count > 0:  # Step 1 above.
            count -= 1
            # select random non-zero point
            index = random[count]
            x = x_idxs[index]
            y = y_idxs[index]

            # if previously eliminated, skip.
            # Note - this will become slow when we have removed many pixels
            # from the mask.
            if not mask[y, x]:
                continue

            value = 0
            max_value = threshold - 1  # Max value in accumulator.
            max_theta_idx = -1

            # Apply Hough transform on point (step 2 above).
            for j in range(nthetas):
                accum_idx = round((ctheta[j] * x + stheta[j] * y)) + offset
                accum[accum_idx, j] += 1
                value = accum[accum_idx, j]
                if value > max_value:
                    max_value = value
                    max_theta_idx = j
            if max_value < threshold:  # Step 4 above.
                continue

            # From the random point (x, y), walk in opposite directions and
            # find line beginning and end (step 5 above).
            line_sin = stheta[max_theta_idx]
            line_cos = ctheta[max_theta_idx]
            a = -line_sin
            b = line_cos
            x0 = x
            y0 = y
            # calculate gradient of walks using fixed point math
            xflag = fabs(a) > fabs(b)
            if xflag:
                if a > 0:
                    dx0 = 1
                else:
                    dx0 = -1
                dy0 = round(b * (1 << shift) / fabs(a))
                y0 = (y0 << shift) + (1 << (shift - 1))
            else:
                if b > 0:
                    dy0 = 1
                else:
                    dy0 = -1
                dx0 = round(a * (1 << shift) / fabs(b))
                x0 = (x0 << shift) + (1 << (shift - 1))

            # pass 1: walk the line, merging lines less than specified gap
            # length (step 5 continued).
            for k in range(2):
                gap = 0
                px = x0
                py = y0
                dx = dx0
                dy = dy0
                if k > 0:  # Walk in opposite direction.
                    dx = -dx
                    dy = -dy
                while 1:
                    if xflag:
                        x1 = px
                        y1 = py >> shift
                    else:
                        x1 = px >> shift
                        y1 = py
                    # check when line exits image boundary
                    if x1 < 0 or x1 >= width or y1 < 0 or y1 >= height:
                        break
                    gap += 1
                    if mask[y1, x1]:  # Hit remaining pixel, continue line.
                        gap = 0
                        line_end[2*k] = x1
                        line_end[2*k + 1] = y1
                    elif gap > line_gap:  # Gap to here too large, end line.
                        break
                    px += dx
                    py += dy

            # Confirm line length is sufficient.
            good_line = (labs(line_end[3] - line_end[1]) >= line_length or
                         labs(line_end[2] - line_end[0]) >= line_length)
            if not good_line:
                continue

            # pass 2: walk the line again and reset accumulator and mask
            # Steps 6 and 7 above.
            for k in range(2):
                px = x0
                py = y0
                dx = dx0
                dy = dy0
                if k > 0:
                    dx = -dx
                    dy = -dy
                while True:
                    if xflag:
                        x1 = px
                        y1 = py >> shift
                    else:
                        x1 = px >> shift
                        y1 = py
                    if mask[y1, x1]:  # Remaining point at this location.
                        mask[y1, x1] = 0  # Remove.
                        for j in range(nthetas):  # Reset accumulator.
                            rho = ctheta[j] * x1 + stheta[j] * y1
                            rho_idx = <int>round(rho) + offset
                            accum[rho_idx, j] -= 1
                    # Exit when the point is the line end.
                    if x1 == line_end[2*k] and y1 == line_end[2*k + 1]:
                        break
                    px += dx
                    py += dy

            # Add line to the result (step 8 above).
            lines[nlines, 0, 0] = line_end[0]
            lines[nlines, 0, 1] = line_end[1]
            lines[nlines, 1, 0] = line_end[2]
            lines[nlines, 1, 1] = line_end[3]
            nlines += 1
            if nlines >= lines_max:
                break

    PyMem_Free(line_end)
    return [((line[0, 0], line[0, 1]), (line[1, 0], line[1, 1]))
            for line in lines[:nlines]]
